package sketch;

public final class Comparison extends SketchObject implements Comparable<Comparison> {
	
	public Comparison(){}
	
	public Comparison(int[] buffer){
		this(buffer, null);
	}
	
	public Comparison(Sketch b){
		this(null, b);
	}
	
	public Comparison(int[] buffer, Sketch b){
		if(buffer!=null){
			hits=buffer[0];
			minIndex=buffer[1];
			minLen=buffer[2];
			maxLen=buffer[3];
		}
		
		if(b!=null){
			name0=b.name0();
			taxName=b.taxName();
			taxID=b.taxID;
			imgID=b.imgID;
			genomeSizeBases=b.genomeSizeBases;
			genomeSizeKmers=b.genomeSizeKmers;
			genomeSizeEstimate=b.genomeSizeEstimate();
		}
	}

	@Override
	public int compareTo(Comparison b) {
		
		{
			float pa=pivot(), pb=b.pivot();
			if(pa>pb){
				return 1;
			}else if (pa<pb){
				return -1;
			}
		}
		
		int x=hits-b.hits;
		if(x!=0){return x;}
		x=b.minIndex-minIndex;
		if(x!=0){return x;}
		x=b.minLen-minLen;
		if(x!=0){return x;}
		x=b.maxLen-maxLen;
		if(x!=0){return x;}
		x=taxID-b.taxID;
		if(x!=0){return x;}
		if(name0!=null && b.name0!=null){
			return name0.compareTo(b.name0);
		}
		if(taxName!=null && b.taxName!=null){
			return taxName.compareTo(b.taxName);
		}
		return 0;
	}
	
	public boolean equals(Object b){
		if(b==null || b.getClass()!=this.getClass()){return false;}
		return compareTo((Comparison)b)==0;
	}
	
	public float idWeighted(){
		return hits/(float)minIndex;
	}
	
	public float idMin(){
		return hits/(float)minLen;
	}
	
	public float idMax(){
		return hits/(float)maxLen;
	}
	
	public float pivot(){
		return idWeighted()*hits;
	}
	
	public String taxName, name0;
	public int taxID;
	public long imgID;
	public long genomeSizeBases;
	public long genomeSizeKmers;
	public long genomeSizeEstimate;
	
	public int hits;
	public int minIndex;
	public int minLen;
	public int maxLen;
	
}
